// Copyright 2017 Google Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
////////////////////////////////////////////////////////////////////////////////

package com.google.cloud.crypto.tink.subtle;

import com.google.api.services.cloudkms.v1beta1.CloudKMS;
import com.google.api.services.cloudkms.v1beta1.model.DecryptRequest;
import com.google.api.services.cloudkms.v1beta1.model.DecryptResponse;
import com.google.api.services.cloudkms.v1beta1.model.EncryptRequest;
import com.google.api.services.cloudkms.v1beta1.model.EncryptResponse;
import com.google.cloud.crypto.tink.Aead;
import java.io.IOException;
import java.security.GeneralSecurityException;
import java.util.concurrent.Future;

/**
 * This primitive forwards encryption/decryption requests to a key in Google Cloud KMS.
 * As of March 2017, Google Cloud KMS supports only AES-256-GCM keys.
 */
public final class GoogleCloudKmsAead implements Aead {

  /**
   * This client knows how to talk to Google Cloud KMS.
   */
  private final CloudKMS kmsClient;

  // The location of a CryptoKey in Google Cloud KMS.
  // Valid values have this format: projects/*/locations/*/keyRings/*/cryptoKeys/*.
  // See https://cloud.google.com/kms/docs/object-hierarchy.
  private final String kmsKeyUri;

  public GoogleCloudKmsAead(CloudKMS kmsClient, String kmsKeyUri) {
    this.kmsClient = kmsClient;
    this.kmsKeyUri = kmsKeyUri;
  }

  @Override
  public byte[] encrypt(final byte[] plaintext, final byte[] aad) throws GeneralSecurityException {
    try {
      EncryptRequest request = new EncryptRequest()
          .encodePlaintext(plaintext)
          .encodeAdditionalAuthenticatedData(aad);
      EncryptResponse response = this.kmsClient.projects().locations().keyRings().cryptoKeys()
          .encrypt(this.kmsKeyUri, request)
          .execute();
      return response.decodeCiphertext();
    } catch (IOException e) {
      throw new GeneralSecurityException("encryption failed");
    }

  }

  @Override
  public byte[] decrypt(final byte[] ciphertext, final byte[] aad)
      throws GeneralSecurityException {
    try {
      DecryptRequest request = new DecryptRequest()
          .encodeCiphertext(ciphertext)
          .encodeAdditionalAuthenticatedData(aad);
      DecryptResponse response = this.kmsClient.projects().locations().keyRings().cryptoKeys()
          .decrypt(this.kmsKeyUri, request)
          .execute();
      return response.decodePlaintext();
    } catch (IOException e) {
      throw new GeneralSecurityException("decryption failed");
    }

  }

  @Override
  public Future<byte[]> asyncEncrypt(final byte[] plaintext, final byte[] aad)
      throws GeneralSecurityException {
    throw new GeneralSecurityException("Not Implemented Yet");
  }

  @Override
  public Future<byte[]> asyncDecrypt(final byte[] ciphertext, final byte[] aad)
      throws GeneralSecurityException {
    throw new GeneralSecurityException("Not Implemented Yet");
  }
}
